"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ScriptCat = void 0;
const cron_1 = require("cron");
const gm_1 = require("./gm/gm");
const utils_1 = require("./utils");
const tough_cookie_1 = require("tough-cookie");
const log4js_1 = require("log4js");
const axios_1 = __importDefault(require("axios"));
(0, log4js_1.configure)({
    appenders: { console: { type: 'console' } },
    categories: { default: { appenders: ['console'], level: 'info' } },
});
class ScriptCat {
    constructor() {
        this.logger = (0, log4js_1.getLogger)();
    }
    Run(script, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const meta = (0, utils_1.parseMetadata)(script);
            if (!meta) {
                throw new Error('parse metadata');
            }
            const gmContext = this.buildGmCtx(options);
            const func = yield this.compile(meta, script);
            const ctx = this.compileContext(gmContext, meta);
            if (meta['background']) {
                return this.runOnce(ctx, func);
            }
            //TODO: 定时脚本的管理
            return new Promise((resolve) => {
                if (!meta['crontab']) {
                    throw new Error('not support');
                }
                const unit = meta['crontab'][0].split(' ');
                if (unit.length == 5) {
                    unit.unshift('0');
                }
                new cron_1.CronJob(unit.join(' '), () => {
                    void this.runOnce(ctx, func);
                }, null, true);
            });
        });
    }
    RunOnce(script, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const meta = (0, utils_1.parseMetadata)(script);
            if (!meta) {
                throw new Error('parse metadata');
            }
            const gmContext = this.buildGmCtx(options);
            const func = yield this.compile(meta, script);
            const ctx = this.compileContext(gmContext, meta);
            return this.runOnce(ctx, func);
        });
    }
    cookieJar(cookie) {
        const jar = new tough_cookie_1.CookieJar();
        for (let i = 0; i < cookie.length; i++) {
            const cookies = cookie[i].cookies;
            if (!cookies) {
                continue;
            }
            for (let n = 0; n < cookies.length; n++) {
                const cookie = cookies[n];
                let u = cookie.secure ? 'https://' : 'http://';
                if (cookie.domain.startsWith('.')) {
                    cookie.domain = cookie.domain.substring(1);
                    u += 'www.' + cookie.domain;
                }
                else {
                    u += cookie.domain;
                }
                u += '/';
                const c = new tough_cookie_1.Cookie({
                    key: cookie.name,
                    value: cookie.value,
                    expires: cookie.expirationDate
                        ? new Date(cookie.expirationDate * 1000)
                        : undefined,
                    maxAge: cookie.expirationDate ? undefined : 'Infinity',
                    // maxAge,extensions,creation,creationIndex,pathIsDefault,lastAccessed,sameSite
                    domain: cookie.domain,
                    path: cookie.path,
                    secure: cookie.secure,
                    httpOnly: cookie.httpOnly,
                    hostOnly: cookie.hostOnly,
                });
                jar.setCookieSync(c, u);
            }
        }
        return jar;
    }
    buildGmCtx(options) {
        let jar;
        if (options && options.cookies) {
            jar = this.cookieJar(options.cookies);
        }
        return new gm_1.GMContext((0, log4js_1.getLogger)('gm'), jar, options === null || options === void 0 ? void 0 : options.values);
    }
    runOnce(context, func) {
        return new Promise((resolve, reject) => {
            const sandbox = this.sandboxThis(context);
            const result = func.apply(sandbox, [sandbox]);
            if (!(result instanceof Promise)) {
                return resolve(this.resultToString(result));
            }
            void result
                .then((result) => {
                resolve(this.resultToString(result));
            })
                .catch((err) => {
                reject(err);
            });
        });
    }
    // 沙盒
    sandboxThis(context) {
        let global = {};
        return new Proxy(context, {
            get(_, name) {
                switch (name) {
                    case 'window':
                    case 'self':
                    case 'global':
                    case 'globalThis':
                        return global;
                }
                if (typeof name == 'string' && name !== 'undefined') {
                    if (context[name]) {
                        // eslint-disable-next-line @typescript-eslint/no-unsafe-return
                        return context[name];
                    }
                    if (global[name] !== undefined) {
                        if (typeof global[name] === 'function' &&
                            !global[name].prototype) {
                            return global[name].bind(global);
                        }
                        // eslint-disable-next-line @typescript-eslint/no-unsafe-return
                        return global[name];
                    }
                }
                return undefined;
            },
            has(_, name) {
                switch (name) {
                    case 'window':
                    case 'self':
                    case 'global':
                    case 'globalThis':
                        return true;
                }
                if (typeof name === 'string') {
                    return context[name] || global[name] ? true : false;
                }
                return false;
            },
            set(_, name, val) {
                switch (name) {
                    case 'window':
                    case 'self':
                    case 'global':
                    case 'globalThis':
                        global = val;
                        return true;
                }
                context[name] = val;
                return true;
            },
        });
    }
    resultToString(result) {
        if (typeof result == 'string') {
            return result;
        }
        return JSON.stringify(result);
    }
    compile(meta, script) {
        return __awaiter(this, void 0, void 0, function* () {
            // 加载@require的js
            if (meta['require']) {
                for (let i = 0; i < meta['require'].length; i++) {
                    try {
                        const resp = yield (0, axios_1.default)(meta['require'][i]);
                        script = resp.data + '\n' + script;
                    }
                    catch (e) { }
                }
            }
            const func = (new Function('context', 'with (context) return ((context)=>{\n' + script + '\n})(context)'));
            return Promise.resolve(func);
        });
    }
    // 编译上下文this
    compileContext(gmContext, meta) {
        const ret = {};
        if (meta['grant']) {
            for (let i = 0; i < meta['grant'].length; i++) {
                const func = gmContext.GmFunc(meta['grant'][i]).bind(gmContext);
                if (func) {
                    ret[meta['grant'][i]] = func;
                }
            }
        }
        return ret;
    }
}
exports.ScriptCat = ScriptCat;
